from typing import Union, Callable
from numbers import Number
from warnings import warn

from computegraph import ComputeGraph
from computegraph.types import GraphObject, Function
import jax

from numpy import ndarray

from summer2.parameters import Time

from .util import enforce_graph_array, piecewise_constant
from . import interpolate

GraphArrayLike = Union[jax.Array, ndarray, list]


def get_linear_interpolation_function(
    x_pts: GraphArrayLike, y_pts: GraphArrayLike, x_axis=Time
) -> Function:
    """Build a linear interpolator over the x and y points supplied

    Args:
        x_pts: Array of x points over which to interpolate
        y_pts: Values at f(x) for x_pts, over which to interpolate
        x_axis: First argument to this function; defaults to Time.

    Returns:
        Resultant GraphObject Function
    """
    x_pts_go = enforce_graph_array(x_pts)
    y_pts_go = enforce_graph_array(y_pts)
    xscale = Function(interpolate.get_scale_data, [x_pts_go])
    yscale = Function(interpolate.get_scale_data, [y_pts_go])

    return Function(interpolate.interpolate_linear, [x_axis, xscale, yscale])


def get_sigmoidal_interpolation_function(
    x_pts: GraphArrayLike, y_pts: GraphArrayLike, x_axis=Time, curvature=16.0
) -> Function:
    """Build a piecewise sigmoidal curve interpolator over the x and y points supplied

    Args:
        x_pts: Array of x points over which to interpolate
        y_pts: Values at f(x) for x_pts, over which to interpolate
        x_axis: First argument to this function; defaults to Time.

        curvature (optional): Curvature of the sigmoid. Defaults to 16.0.

    Returns:
        Function: _description_
    """
    x_pts_go = enforce_graph_array(x_pts)
    y_pts_go = enforce_graph_array(y_pts)
    xscale = Function(interpolate.get_scale_data, [x_pts_go])
    yscale = Function(interpolate.get_scale_data, [y_pts_go])

    curve = interpolate.build_sigmoidal_multicurve(curvature=curvature)

    return Function(curve, [x_axis, xscale, yscale])


def get_piecewise_scalar_function(
    breakpoints: GraphArrayLike, values: GraphArrayLike, x_axis=Time
) -> Function:
    warn(
        "This method is deprecated and scheduled for removal, use get_piecewise_function instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_piecewise_function(breakpoints, values, x_axis)


def get_piecewise_function(
    breakpoints: GraphArrayLike, values: GraphArrayLike, x_axis=Time
) -> Function:
    """Create a Function object to be evaluated in the {x} domain, returning
    a value selected from {values}, the index of which is the x-bounds
    described by {breakpoints}.

    By default x is summer2 Time (time-varying function)

    Args:
        breakpoints: Breakpoints to which x is compared
        values: Array-like of length len(breakpoints) + 1
        x: GraphObject supplying the x value.  Defaults to Time.

    Returns:
        The resulting wrapped Function object
    """
    breakpoints = enforce_graph_array(breakpoints)
    values = enforce_graph_array(values)
    return Function(piecewise_constant, (x_axis, breakpoints, values))


def _tfunc(f: Function):
    return lambda t, parameters: f(model_variables={"time": t}, parameters=parameters)["out"]


def get_time_callable(f: Function, jit_compile=True) -> Callable:
    """For a Function f which takes Time and Parameters as arguments,
    return a python callable accepting either a number or an an array,
    and a dictionary of parameter values.

    E.g
    returned_func(1.0, parameters={"scale": 0.5})

    Args:
        f: The Time Function
        jit_compile (optional): JIT the function

    Returns:
        The resulting function
    """
    cg = ComputeGraph(f)
    tf = _tfunc(cg.get_callable())
    if jit_compile:
        tf = jax.jit(tf)
    vmapped = jax.vmap(tf, in_axes=(0, None))
    if jit_compile:
        vmapped = jax.jit(vmapped)

    def wrapped(t, parameters=None):
        if isinstance(t, Number):
            return tf(t, parameters)
        else:
            return vmapped(t, parameters)

    return wrapped



"""Builder functions for interpolating data - replaces scale_up.py

Note: For most user applications, refer to the get_*_function wrappers in time.py

build_static_sigmoidal_multicurve is the equivalent of scale_up_function(method=4),
with build_sigmoidal_multicurve producing a dynamic (parameterizable) version of this

"""

from jax import Array, lax, numpy as jnp, lax

from .util import binary_search_ge
from collections import namedtuple

# Define datatypes used within interpolation functions
# These carry intermediary compute-once values like ranges/bounds

InterpolatorScaleData = namedtuple("InterpolatorScaleData", ["points", "ranges", "bounds"])


def _uncorrected_sigmoid(x: float, curvature: float) -> float:
    """Return a sigmoid of x (assumed to be between 0.0 and 1.0),
    whose shape and range depends on curvature

    Args:
        x: Input to the function
        curvature: Degree of sigmoidal flattening - 1.0 is linear, higher values increase smoothing

    Returns:
        Output of sigmoidal curve function
    """
    arg = curvature * (0.5 - x)
    return 1.0 / (1.0 + jnp.exp(arg))


def make_norm_sigmoid(curvature: float) -> callable:
    """
    Build a sigmoid function with fixed curvature, whose output is normalized to (0.0,1.0) across
    the (0.0,1.0) input range
    Args:
        curvature: See _uncorrected_sigmoid

    Returns:
        The normalized sigmoid function
    """
    offset = _uncorrected_sigmoid(0.0, curvature)
    scale = 1.0 / (1.0 - (offset * 2.0))

    def sig(x):
        return (_uncorrected_sigmoid(x, curvature) - offset) * scale

    return sig


def build_sigmoidal_multicurve(curvature=16.0) -> callable:
    """Build a sigmoidal smoothing function across points specified by
    x_points; the returned function takes the y values as arguments in
    form of a scale_data dict with keys [min, max, values, ranges]

    Args:
        x_points: x values to interpolate across
        curvature: Sigmoidal curvature.  Default produces the same behaviour as the old
                   scale_up_function

    Returns:
        The multisigmoidal function of (x, scale_data)
    """

    sig = make_norm_sigmoid(curvature)

    def _get_sigmoidal_curve_at_x(x, xdata, ydata):
        # idx = sum(t >= x_points) - 1
        idx = binary_search_ge(x, xdata.points)

        offset = x - xdata.points[idx]
        relx = offset / xdata.ranges[idx]
        rely = sig(relx)
        return ydata.points[idx] + (rely * ydata.ranges[idx])

    def interpolate_sigmoidal(t: float, xdata: InterpolatorScaleData, ydata: InterpolatorScaleData):
        # Branch on whether t is in bounds
        bounds_state = sum(t > xdata.bounds)
        branches = [
            lambda _, __, ___: ydata.bounds[0],
            _get_sigmoidal_curve_at_x,
            lambda _, __, ___: ydata.bounds[1],
        ]
        return lax.switch(bounds_state, branches, t, xdata, ydata)

    return interpolate_sigmoidal


#
# Linear Interpolation
#


def _get_linear_curve_at_x(
    x: float, xdata: InterpolatorScaleData, ydata: InterpolatorScaleData
) -> float:
    """Inner function for interpolate_linear

    Args:
        x: x position (usually Time)
        xdata: Scale data for x ('index') points
        ydata: Scale data for y ('value') points

    Returns:
        The interpolated output value
    """
    idx = binary_search_ge(x, xdata.points)

    offset = x - xdata.points[idx]
    relx = offset / xdata.ranges[idx]
    return ydata.points[idx] + (relx * ydata.ranges[idx])


def interpolate_linear(t: float, xdata: InterpolatorScaleData, ydata: InterpolatorScaleData):
    # Branch on whether t is in bounds
    bounds_state = sum(t > xdata.bounds)
    branches = [
        lambda _, __, ___: ydata.bounds[0],
        _get_linear_curve_at_x,
        lambda _, __, ___: ydata.bounds[1],
    ]
    return lax.switch(bounds_state, branches, t, xdata, ydata)


def get_scale_data(points: Array) -> InterpolatorScaleData:
    """
    Precompute ranges (diffs) and bounds (left and right extrema) for a set of data to be used in
    a scaling function such as that produced by build_sigmoidal_multicurve.  The onus is on the
    caller of this function to ensure they are the length expected by the target callee
    """
    ranges = jnp.diff(points)
    lpoint = points[0]
    rpoint = points[-1]

    # data = {"min": ymin, "max": ymax, "values": points, "ranges": ranges}
    return InterpolatorScaleData(points, ranges, jnp.array([lpoint, rpoint]))



from . import time
from . import util
from . import interpolate

from .time import (
    get_piecewise_scalar_function,
    get_linear_interpolation_function,
    get_sigmoidal_interpolation_function,
)



"""Implementation of CompartmentalModel and ModelBackend internals in Jax

This is a mess right now!
"""

from dataclasses import dataclass

import numpy as np

from jax import lax, numpy as jnp

from summer2.runner.jax import ode

# from jax.experimental import ode
from summer2.runner.jax import solvers

from summer2.adjust import Overwrite

from summer2.runner import ModelBackend

from summer2.solver import SolverType, SolverArgs

from .stratify import get_calculate_initial_pop
from .derived_outputs import build_derived_outputs_runner


def clean_compartments(compartment_values: jnp.ndarray):
    return jnp.where(compartment_values < 0.0, 0.0, compartment_values)


def get_force_of_infection(
    strain_infectious_values: jnp.array,
    strain_compartment_infectiousness: jnp.array,
    strain_category_indexer: jnp.array,
    mixing_matrix: jnp.array,
    category_populations: jnp.array,
) -> dict:
    """
    Calculate the force of infection for both frequency-dependent and density-dependent transmission assumptions.
    Considers only one strain at this stage, with the strain logic sitting outside of this function.

    Args:
        strain_infectious_values: Vector of compartment size values for the infectious compartments (relevant to the strain considered)
        strain_compartment_infectiousness: Vector of infectiousness scaling values with same length as strain_infectious_values
        strain_category_indexer: Indexer with each row containing the indices of compartments relevant to a mixing category
        mixing_matrix: The square mixing matrix with dimensions equal to the number of mixing categories
            The columns of the matrix represent the infecting categories and the rows the infected categories
        category_populations: Vector of the population sizes with length equal to the number of mixing categories
    """

    infected_values = strain_infectious_values * strain_compartment_infectiousness
    infectious_populations = jnp.sum(infected_values[strain_category_indexer], axis=-1)
    infection_density = mixing_matrix @ infectious_populations
    category_prevalence = infectious_populations / category_populations
    infection_frequency = mixing_matrix @ category_prevalence

    return {"infection_density": infection_density, "infection_frequency": infection_frequency}


def build_get_infectious_multipliers(runner, debug=False):
    population_cat_indexer = jnp.array(runner._population_category_indexer)

    # FIXME: We are hardcoding this for frequency only right now
    infect_proc_type = runner._infection_process_type

    # We use this elsewhere to enable testing simple models without an infection process
    if infect_proc_type is None:
        return None

    if infect_proc_type == "both":
        raise NotImplementedError("No support for mixed infection frequency/density")

    # FIXME: This could desparately use a tidy-up - all the indexing is a nightmare

    def get_infectious_multipliers(
        time, compartment_values, cur_graph_outputs, compartment_infectiousness
    ):
        infection_frequency = {}
        infection_density = {}

        full_multipliers = jnp.ones(len(runner.infectious_flow_indices))

        mixing_matrix = cur_graph_outputs["mixing_matrix"]
        category_populations = compartment_values[population_cat_indexer].sum(axis=1)

        per_strain_out = {}

        for strain_idx, strain in enumerate(runner.model._disease_strains):
            strain_compartment_infectiousness = compartment_infectiousness[strain]
            strain_infectious_idx = runner._strain_infectious_indexers[strain]
            strain_category_indexer = runner._strain_category_indexers[strain]

            strain_infectious_values = compartment_values[strain_infectious_idx]
            strain_values = get_force_of_infection(
                strain_infectious_values,
                strain_compartment_infectiousness,
                strain_category_indexer,
                mixing_matrix,
                category_populations,
            )
            infection_frequency[strain] = strain_values["infection_frequency"]
            infection_density[strain] = strain_values["infection_density"]

            if infect_proc_type == "freq":
                strain_ifect = strain_values["infection_frequency"]
            elif infect_proc_type == "dens":
                strain_ifect = strain_values["infection_density"]

            if debug:
                per_strain_out[strain] = strain_ifect

            # FIXME: So we produce strain infection values _per category_
            # (ie not all model compartments, not all infectious compartments)
            # We need to rebroadcast these back out the appropriate compartments somehow
            # ... hence the following bit of weird double-dip indexing

            # _infect_strain_lookup_idx is an array of size (num_infectious_comps), whose values
            # are the strain_idx of the strain to whom they belong

            strain_ifectcomp_mask = runner._infect_strain_lookup_idx == strain_idx

            # _infect_cat_lookup_idx is an array of size (num_infectious_comps), whose values
            # are the mixing category to which they belong (and therefore an index into
            # the values returned in strain_ifect above)

            # Get the strain infection values broadcast to num_infectious_comps
            strain_ifect_bcast = strain_ifect[runner._infect_cat_lookup_idx]

            # full_multipliers is the length of the infectious compartments - ie the same as the
            # above mask
            # full_multipliers = full_multipliers.at[strain_ifectcomp_mask].mul(
            #    strain_ifect_bcast[strain_ifectcomp_mask]
            # )

            full_multipliers = full_multipliers.at[strain_ifectcomp_mask].set(
                full_multipliers[strain_ifectcomp_mask] * strain_ifect_bcast[strain_ifectcomp_mask]
            )

        if debug:
            return full_multipliers, per_strain_out
        else:
            return full_multipliers

    return get_infectious_multipliers


def build_get_flow_weights(runner: ModelBackend):
    m = runner.model

    dag_keys = list(m.graph.dag)
    mvars = [k for k in dag_keys if k.startswith("model_variables.")]

    if len(mvars):
        tv_keys = list(m.graph.filter(sources=mvars).dag)
        tv_flow_map = {
            k: m._flow_key_map[k] for k in set(m._flow_key_map).intersection(set(tv_keys))
        }
    else:
        tv_flow_map = {}

    def get_flow_weights(cur_graph_outputs, static_flow_weights):
        flow_weights = jnp.copy(static_flow_weights)

        for k, v in tv_flow_map.items():
            val = cur_graph_outputs[k]
            flow_weights = flow_weights.at[v].set(val)

        return flow_weights

    return get_flow_weights


def _build_calc_computed_values(runner):
    def calc_computed_values(compartment_vals, time, parameters):
        model_variables = {"compartment_values": compartment_vals, "time": time}

        computed_values = runner.computed_values_runner(
            parameters=parameters, model_variables=model_variables
        )

        return computed_values

    return calc_computed_values


def build_get_flow_rates(runner, ts_graph_func, get_infectious_multipliers=None, debug=False):
    # calc_computed_values = build_calc_computed_values(runner)
    get_flow_weights = build_get_flow_weights(runner)

    infect_proc_type = runner._infection_process_type
    # if infect_proc_type:
    #    get_infectious_multipliers = build_get_infectious_multipliers(runner)

    population_idx = np.array(runner.population_idx)
    infectious_flow_indices = jnp.array(runner.infectious_flow_indices)

    def get_flow_rates(compartment_values: jnp.array, time, static_graph_vals, model_data):
        compartment_values = clean_compartments(compartment_values)

        sources = {
            "model_variables": {"time": time, "compartment_values": compartment_values},
            "static_inputs": static_graph_vals,
        }

        ts_graph_vals = ts_graph_func(**sources)

        # JITTED
        flow_weights = get_flow_weights(ts_graph_vals, model_data["static_flow_weights"])

        populations = compartment_values[population_idx]

        # Update for special cases (population-independent and CrudeBirth)
        if runner._has_non_pop_flows:
            populations = populations.at[runner._non_pop_flow_idx].set(1.0)
        if runner._has_crude_birth:
            populations = populations.at[runner._crude_birth_idx].set(compartment_values.sum())

        flow_rates = flow_weights * populations

        # Calculate infection flows
        if infect_proc_type:
            infect_mul = get_infectious_multipliers(
                time,
                compartment_values,
                ts_graph_vals,
                model_data["compartment_infectiousness"],
            )
            # flow_rates = flow_rates.at[infectious_flow_indices].mul(infect_mul)
            flow_rates = flow_rates.at[infectious_flow_indices].set(
                flow_rates[infectious_flow_indices] * infect_mul
            )
            # ReplacementBirthFlow depends on death flows already being calculated; update here
        if runner._has_replacement:
            # Only calculate timestep_deaths if we use replacement, it's expensive...
            # if len(runner.death_flow_indices):
            _timestep_deaths = flow_rates[runner.death_flow_indices].sum()
            # else:
            #    _timestep_deaths = 0.0
            # flow_rates = flow_rates.at[runner._replacement_flow_idx].set(0.0)  # _timestep_deaths)
            flow_rates = flow_rates.at[runner._replacement_flow_idx].set(
                flow_rates[runner._replacement_flow_idx] * _timestep_deaths
            )

        if debug:
            return flow_rates, ts_graph_vals["computed_values"], ts_graph_vals
        else:
            return flow_rates, ts_graph_vals["computed_values"]

    return get_flow_rates


def build_get_compartment_rates(runner):
    accum_maps = get_accumulation_maps(runner)

    def get_compartment_rates(compartment_values, flow_rates):
        comp_rates = jnp.zeros_like(compartment_values)

        for flow_src, comp_target in accum_maps["positive"]:
            comp_rates = comp_rates.at[comp_target].add(flow_rates[flow_src])
        for flow_src, comp_target in accum_maps["negative"]:
            comp_rates = comp_rates.at[comp_target].add(-flow_rates[flow_src])

        return comp_rates

    return get_compartment_rates


def build_get_rates(runner, ts_graph_func):
    get_infectious_multipliers = build_get_infectious_multipliers(runner)
    get_flow_rates = build_get_flow_rates(runner, ts_graph_func, get_infectious_multipliers)
    get_flow_rates_debug = build_get_flow_rates(
        runner, ts_graph_func, get_infectious_multipliers, True
    )
    get_compartment_rates = build_get_compartment_rates(runner)

    def get_rates(compartment_values, time, static_graph_vals, model_data):
        flow_rates, _ = get_flow_rates(compartment_values, time, static_graph_vals, model_data)
        comp_rates = get_compartment_rates(compartment_values, flow_rates)

        return flow_rates, comp_rates

    def get_rates_debug(compartment_values, time, static_graph_vals, model_data):
        flow_rates, _, cur_ts_vals = get_flow_rates_debug(
            compartment_values, time, static_graph_vals, model_data
        )
        comp_rates = get_compartment_rates(compartment_values, flow_rates)

        return flow_rates, comp_rates, cur_ts_vals

    return {
        "get_flow_rates": get_flow_rates,
        "get_rates": get_rates,
        "get_rates_debug": get_rates_debug,
        "get_infectious_multipliers": get_infectious_multipliers,
    }


def get_accumulation_maps(runner):
    pos_map = [mflow for mflow in runner._pos_flow_map]
    neg_map = [mflow for mflow in runner._neg_flow_map]

    def peel_flow_map(flow_map):
        targets = []
        src_idx = []
        remainder = []
        for src_flow, target in flow_map:
            if target not in targets:
                targets.append(target)
                src_idx.append(src_flow)
            else:
                remainder.append((src_flow, target))
        return np.array(src_idx), np.array(targets), remainder

    def recurse_unpeel(flow_map):
        remainder = flow_map
        full_map = []
        while len(remainder) > 0:
            sources, targets, remainder = peel_flow_map(remainder)
            full_map.append((sources, targets))
        return full_map

    return {"positive": recurse_unpeel(pos_map), "negative": recurse_unpeel(neg_map)}


def build_get_compartment_infectiousness(model):
    """
    Build a Jax function to return the compartment infectiousness (for all compartments),
    of the strain specified by strain
    """

    # This is run during prepare_dynamic
    # i.e. it is done once at the start of a model run, but
    # is parameterized (non-structural)
    def get_compartment_infectiousness(static_graph_values):
        # Find the infectiousness multipliers for each compartment being implemented in the model.
        compartment_infectiousness = jnp.ones(len(model.compartments))

        # Apply infectiousness adjustments
        for strat in model._stratifications:
            for comp_name, adjustments in strat.infectiousness_adjustments.items():
                for stratum, adjustment in adjustments.items():
                    if adjustment:
                        is_overwrite = isinstance(adjustment, Overwrite)
                        adj_value = static_graph_values[adjustment.param._graph_key]
                        adj_comps = model.get_matching_compartments(
                            comp_name, {strat.name: stratum}
                        )
                        for c in adj_comps:
                            if is_overwrite:
                                compartment_infectiousness = compartment_infectiousness.at[
                                    c.idx
                                ].set(adj_value)
                            else:
                                orig_value = compartment_infectiousness[c.idx]
                                compartment_infectiousness = compartment_infectiousness.at[
                                    c.idx
                                ].set(adj_value * orig_value)

        strain_comp_inf = {}

        for strain in model._disease_strains:
            if "strain" in model.stratifications:
                strain_filter = {"strain": strain}
            else:
                strain_filter = {}

            # _Must_ be ordered here
            strain_infect_comps = model.query_compartments(
                strain_filter, tags="infectious", as_idx=True
            )

            strain_comp_inf[strain] = compartment_infectiousness[strain_infect_comps]

        return strain_comp_inf

    return get_compartment_infectiousness


# Returned by the one_step runner function - the complete internal state
# of the model at a given timestep
# Does not include derived outputs - these are a post-process
@dataclass
class StepResults:
    flow_rates: jnp.array
    comp_rates: jnp.array
    comp_vals: jnp.array
    static_graph_vals: dict
    ts_graph_vals: dict
    initial_population: jnp.array
    model_data: dict
    infectious_multipliers: jnp.array
    infect_mul_per_strain: dict


def build_run_model(
    runner,
    base_params=None,
    dyn_params=None,
    solver=None,
    solver_args=None,
    derived_outputs=None,
    include_full_outputs=True,
):
    if dyn_params is None:
        dyn_params = runner.model.get_input_parameters()

    dyn_params = [f"parameters.{p}" if not p.startswith("parameters.") else p for p in dyn_params]

    # dyn_params may contain parameters only used in derived outputs, that do not show up
    # in the main graph.  These need to be filtered out; they will be computed every time,
    # but it's simpler than trying to squash the graphs together
    model_graph_keys = set(runner.model.graph.dag)
    dyn_params = [k for k in dyn_params if k in model_graph_keys]

    # Graph frozen for all non-calibration parameters
    if base_params is None:
        base_params = {}

    source_inputs = {"parameters": base_params}

    ts_vars = runner.model.graph.query("model_variables")

    dyn_params = set(dyn_params).union(set(ts_vars))

    param_frozen_cg, _ = runner.model.graph.freeze(dyn_params, source_inputs)

    # static_cg = param_frozen_cg.filter(exclude=ts_vars)
    # static_graph_func = static_cg.get_callable()(parameters=base_params)

    timestep_cg, static_cg = param_frozen_cg.freeze(ts_vars)

    timestep_graph_func = timestep_cg.get_callable()
    # timestep_graph_func = timestep_cg.get_callable()
    static_graph_func = static_cg.get_callable()

    rates_funcs = build_get_rates(runner, timestep_graph_func)
    get_rates = rates_funcs["get_rates"]
    get_flow_rates = rates_funcs["get_flow_rates"]
    get_rates_debug = rates_funcs["get_rates_debug"]
    get_infectious_multipliers = rates_funcs["get_infectious_multipliers"]

    # from jax import vmap

    # get_flows_for_outputs = vmap(get_flow_rates, in_axes=(0, 0, None, None), out_axes=(0))

    def get_comp_rates(comp_vals, t, static_graph_vals, model_data):
        return get_rates(comp_vals, t, static_graph_vals, model_data)[1]

    if solver is None or solver == SolverType.SOLVE_IVP:
        solver = SolverType.ODE_INT

    if solver == SolverType.ODE_INT:
        if solver_args is None:
            # Some sensible defaults; faster than
            # the odeint defaults,
            # but accurate enough for our tests
            solver_args = SolverArgs.DEFAULT

        def get_ode_solution(initial_population, times, static_graph_vals, model_data):
            return ode.odeint(
                get_comp_rates,
                initial_population,
                times,
                static_graph_vals,
                model_data,
                **solver_args,
            )

    elif solver == SolverType.RUNGE_KUTTA:

        def get_ode_solution(initial_population, times, static_graph_vals, model_data):
            return solvers.rk4(
                get_comp_rates, initial_population, times, static_graph_vals, model_data
            )

    elif solver == SolverType.EULER:

        def get_ode_solution(initial_population, times, static_graph_vals, model_data):
            return solvers.euler(
                get_comp_rates, initial_population, times, static_graph_vals, model_data
            )

    else:
        raise NotImplementedError("Incompatible SolverType for Jax runner", solver)

    times = jnp.array(runner.model.times)

    calc_initial_pop = get_calculate_initial_pop(runner.model)
    get_compartment_infectiousness = build_get_compartment_infectiousness(runner.model)

    do_cg, calc_derived_outputs = build_derived_outputs_runner(
        runner.model, whitelist=derived_outputs
    )

    m = runner.model
    dag_keys = list(m.graph.dag)

    mvars = [k for k in dag_keys if k.startswith("model_variables.")]

    if len(mvars):
        tv_keys = list(m.graph.filter(sources=mvars).dag)
    else:
        tv_keys = []

    static_flow_map = {k: m._flow_key_map[k] for k in set(m._flow_key_map).difference(set(tv_keys))}

    model_times = jnp.array(m.times)

    # In the case of parameters used by derived outputs that are _not_ supplised in dyn_params,
    # we need to capture these from the original base_params
    do_params = set(
        [v.key for v in m._do_tracker_graph.get_input_variables() if v.source == "parameters"]
    )
    do_base_params = {k: v for k, v in base_params.items() if k in do_params}

    def get_flows_for_outputs(outputs, static_graph_vals, model_data):
        # Empty array to kick-start flow rates from outputs
        flow_rates_full = jnp.empty((len(m.times), len(m.flows)))

        # And likewise for computed values (but as a dict)
        cv_out = {}
        for k in m._computed_values_graph_dict:
            cv_out[k] = jnp.empty((len(m.times),))

        def f(carry, i):
            frate_full, cur_cv = carry
            t = model_times[i]
            frates, cvals = get_flow_rates(outputs[i], t, static_graph_vals, model_data)
            frate_full = frate_full.at[i].set(frates)
            for k, v in cvals.items():
                cur_cv[k] = cur_cv[k].at[i].set(v)
            return (frate_full, cur_cv), None

        (out_flows, out_cv), _ = lax.scan(
            f, (flow_rates_full, cv_out), jnp.array(range(len(m.times)))
        )

        return out_flows, out_cv

    def run_model(parameters):
        static_graph_vals = static_graph_func(parameters=parameters)
        initial_population = calc_initial_pop(static_graph_vals)

        static_flow_weights = jnp.zeros(len(runner.model.flows))
        for k, v in static_flow_map.items():
            val = static_graph_vals[k]
            static_flow_weights = static_flow_weights.at[v].set(val)

        compartment_infectiousness = get_compartment_infectiousness(static_graph_vals)
        model_data = {
            "compartment_infectiousness": compartment_infectiousness,
            "static_flow_weights": static_flow_weights,
        }

        outputs = get_ode_solution(initial_population, times, static_graph_vals, model_data)

        # # Empty array to kick-start flow rates from outputs
        # flow_rates_full = jnp.empty((len(m.times), len(m.flows)))

        # # And likewise for computed values (but as a dict)
        # cv_out = {}
        # for k in m._computed_values_graph_dict:
        #     cv_out[k] = jnp.empty((len(m.times),))

        # def f(carry, i):
        #     frate_full, cur_cv = carry
        #     t = model_times[i]
        #     frates, cvals = get_flow_rates(outputs[i], t, static_graph_vals, model_data)
        #     frate_full = frate_full.at[i].set(frates)
        #     for k, v in cvals.items():
        #         cur_cv[k] = cur_cv[k].at[i].set(v)
        #     return (frate_full, cur_cv), None

        # (out_flows, out_cv), _ = lax.scan(
        #     f, (flow_rates_full, cv_out), jnp.array(range(len(m.times)))
        # )

        out_flows, out_cv = get_flows_for_outputs(outputs, static_graph_vals, model_data)

        model_variables = {"outputs": outputs, "flows": out_flows, "computed_values": out_cv}

        do_full_params = do_base_params.copy()
        do_full_params.update(parameters)

        derived_outputs = calc_derived_outputs(
            parameters=do_full_params, model_variables=model_variables
        )

        # return {"outputs": outputs, "model_data": model_data}
        output_dict = {
            "derived_outputs": derived_outputs,
        }  # "model_data": model_data}
        if include_full_outputs:
            output_dict["outputs"] = outputs

        return output_dict

    ons_get_inf_mul = build_get_infectious_multipliers(runner, True)

    def one_step(parameters: dict = None, t: float = None, comp_vals=None):
        static_graph_vals = static_graph_func(parameters=parameters)

        if t is None:
            t = runner.model.times[0]

        if comp_vals is None:
            comp_vals = calc_initial_pop(static_graph_vals)

        initial_population = comp_vals

        static_flow_weights = jnp.zeros(len(runner.model.flows))
        for k, v in static_flow_map.items():
            val = static_graph_vals[k]
            static_flow_weights = static_flow_weights.at[v].set(val)

        compartment_infectiousness = get_compartment_infectiousness(static_graph_vals)
        model_data = {
            "compartment_infectiousness": compartment_infectiousness,
            "static_flow_weights": static_flow_weights,
        }

        flow_rates, comp_rates, ts_graph_vals = get_rates_debug(
            comp_vals, t, static_graph_vals, model_data
        )
        if get_infectious_multipliers:
            infect_mul, infect_mul_per_strain = ons_get_inf_mul(
                t,
                comp_vals,
                ts_graph_vals,
                compartment_infectiousness,
            )
        else:
            infect_mul, infect_mul_per_strain = None, None

        # return {"outputs": outputs, "model_data": model_data}
        res = {
            "flow_rates": flow_rates,
            "comp_rates": comp_rates,
            "comp_vals": comp_vals + comp_rates,
            "static_graph_vals": static_graph_vals,
            "ts_graph_vals": ts_graph_vals,
            "initial_population": initial_population,
            "model_data": model_data,
            "infectious_multipliers": infect_mul,
            "infect_mul_per_strain": infect_mul_per_strain,
        }
        return StepResults(**res)

    runner_dict = {
        "get_rates": get_rates,
        "get_flow_rates": get_flow_rates,
        "get_comp_rates": get_comp_rates,
        "calc_initial_pop": calc_initial_pop,
        "get_compartment_infectiousness": get_compartment_infectiousness,
        "get_ode_solution": get_ode_solution,
        "calc_derived_outputs": calc_derived_outputs,  # Callable for derived outputs
        "timestep_cg": timestep_cg,  # Computegraph for time-varying processes
        "timestep_graph_func": timestep_graph_func,  # Callable for timestep graph
        "static_cg": static_cg,  # Computegraph for static portion of model
        "static_graph_func": static_graph_func,  # Callable for static graph
        "one_step": one_step,  # Single (euler) step run function
        "derived_outputs_cg": do_cg,  # DerivedOutputs computegraph
        "get_rates_debug": rates_funcs["get_rates_debug"],
    }

    return run_model, runner_dict



from numbers import Number
from collections.abc import Container
from jax import lax, numpy as jnp, Array
from numpy import ndarray
import pandas as pd
import jax

from computegraph.types import GraphObject, Function, Data
from summer2.parameters import Time


def _as_graphobj(x) -> GraphObject:
    if isinstance(x, GraphObject):
        return x
    elif isinstance(x, Number):
        return x
    else:
        jnp_arr = jnp.array(x)
        return Data(jnp_arr)


def enforce_graph_array(x) -> GraphObject:
    # Already a GraphObject
    if isinstance(x, GraphObject):
        return x

    if isinstance(x, pd.Index) or isinstance(x, pd.Series):
        x = x.to_numpy()

    # Numpy array, cast to jax
    if isinstance(x, ndarray):
        x = jnp.array(x)

    # Jax array, return as Data object
    if isinstance(x, Array):
        return Data(x)

    # Something else - let capture_array handle it

    return capture_array(x)


def piecewise_function(x, breakpoints, functions):
    index = sum(x >= breakpoints)
    return lax.switch(index, functions, x)


# All
def piecewise_constant(x, breakpoints, values):
    index = sum(x >= breakpoints)
    return values[index]


def windowed_constant(x: float, value: float, window_start: float, window_length: float):
    breakpoints = jnp.array((window_start, window_start + window_length))
    values = jnp.array((0.0, value, 0.0))
    return piecewise_constant(x, breakpoints, values)


def binary_search_ge(x: float, points: jax.Array) -> int:
    """Find the lowest index of the value within points
    to which x is greater than or equal to, using a binary
    search.

    A value of x lower than min(points) will return 0, and higher
    than max(points) will return len(points)-1

    Args:
        x: Value to find
        points: Array to search

    Returns:
        The index value satisying the above conditions
    """

    def cond(state):
        low, high = state
        return (high - low) > 1

    def body(state):
        low, high = state
        midpoint = (0.5 * (low + high)).astype(int)
        update_upper = x < points[midpoint]
        low = jnp.where(update_upper, low, midpoint)
        high = jnp.where(update_upper, midpoint, high)
        return (low, high)

    low, high = lax.while_loop(cond, body, (0, len(points) - 1))
    return lax.cond(x < points[high], lambda: low, lambda: high)


def capture_dict(**kwargs) -> Function:
    """Create a Function that captures a dictionary

    Returns:
        The GraphObject representation of the dict
    """
    return Function(_capture_dict, kwargs=kwargs)


def _capture_dict(**kwargs):
    return kwargs


def capture_array(arr_obj: list, dtype=None) -> Function:
    """Create a GraphObject representation of an (optionally nested) array type,
    declared as a list

    Args:
        arr_obj: A list of the form consumed by (j)np.array(arr_obj)
        dtype (optional): Datatype of the returned array (inferred by default)

    Returns:
        The GraphObject representation of the array
    """
    ishape = _infer_shape(arr_obj)

    if len(ishape) > 1:
        # arr_obj is a nested listed, which we must flatten to expose args to the graph
        flattened = _flatten_arr_obj(arr_obj)

        def _capture_array_reshape(*args):
            return jnp.array(args, dtype=dtype).reshape(ishape)

        return Function(_capture_array_reshape, args=flattened)
    else:
        # Simple instance of 1d array
        def _capture_array(*args):
            return jnp.array(args, dtype=dtype)

        return Function(_capture_array, args=arr_obj)


def _infer_shape(arr_obj: list, cur_shape=()):
    new_shape = (*cur_shape, len(arr_obj))
    if isinstance(arr_obj[0], list):
        msg = "All subarrays must be lists"
        assert all([isinstance(sub_arr, list) for sub_arr in arr_obj]), msg
        return _infer_shape(arr_obj[0], new_shape)
    return new_shape


def _flatten_arr_obj(arr_obj: list, cur_obj=None):
    if cur_obj is None:
        cur_obj = []
    for v in arr_obj:
        if isinstance(v, list):
            _flatten_arr_obj(v, cur_obj)
        else:
            cur_obj.append(v)
    return cur_obj



from typing import Tuple

import numpy as np
from jax import numpy as jnp

from summer2 import Compartment
from summer2.adjust import Overwrite
import summer2.flows as flows

from summer2.population import get_rebalanced_population
from summer2.utils import clean_compartment_values


class ModelBackend:
    """
    An optimized, but less accessible model runner.
    """

    def __init__(self, model):
        # Compartmental model
        self.model = model

        # Tracks total deaths per timestep for death-replacement birth flows
        self._timestep_deaths = None

        # Set our initial parameters to an empty dict - this is really just to appease tests
        self.parameters = {}

    def prepare_structural(self):
        # FIXME: Redundant
        self._iter_non_function_flows = [(i, f) for i, f in enumerate(self.model.flows)]

        self._build_compartment_category_map()

        self.infectious_flow_indices = np.array(
            [i for i, f in self._iter_non_function_flows if isinstance(f, flows.BaseInfectionFlow)],
            dtype=int,
        )
        self.death_flow_indices = np.array(
            [i for i, f in self._iter_non_function_flows if f.is_death_flow], dtype=int
        )

        # Include dummy values in population_idx to account for Entry flows
        non_func_pops = np.array(
            [f.source.idx if f.source else 0 for i, f in self._iter_non_function_flows], dtype=int
        )

        self.population_idx = non_func_pops

        # Store indices of flows that are not population dependent
        self._non_pop_flow_idx = np.array(
            [
                i
                for i, f in self._iter_non_function_flows
                if (type(f) in (flows.ReplacementBirthFlow, flows.ImportFlow, flows.AbsoluteFlow))
            ],
            dtype=int,
        )
        self._has_non_pop_flows = bool(len(self._non_pop_flow_idx))

        # Crude birth flows use population sum rather than a compartment; store indices here
        self._crude_birth_idx = np.array(
            [i for i, f in self._iter_non_function_flows if type(f) == flows.CrudeBirthFlow],
            dtype=int,
        )
        self._has_crude_birth = bool(len(self._crude_birth_idx))

        # Special indexing required for replacement flows
        self._replacement_flow_idx = np.array(
            [i for i, f in self._iter_non_function_flows if type(f) == flows.ReplacementBirthFlow],
            dtype=int,
        )
        self._has_replacement = bool(len(self._replacement_flow_idx))

        self._precompute_flow_maps()
        self._build_infectious_multipliers_lookup()

    def _build_compartment_category_map(self):
        # Create a matrix that tracks which categories each compartment is in.
        self.num_categories = ncats = len(self.model._mixing_categories)

        # Array that maps compartments to their mixing category
        self._category_lookup = np.empty(len(self.model.compartments), dtype=int)

        all_cat_idx = []
        for i, category in enumerate(self.model._mixing_categories):
            cat_idx = []
            for j, comp in enumerate(self.model.compartments):
                if all(comp.has_stratum(k, v) for k, v in category.items()):
                    cat_idx.append(j)
                    # self._category_matrix[i][j] = 1
                    self._category_lookup[j] = i
            all_cat_idx.append(np.array(cat_idx, dtype=int))

        # Indexer for whole population (compartment_values), that will return a
        # num_cats * (num_comps_per_cat) array that can be summed for category populations
        self._population_category_indexer = pop_cat_idx = np.stack(all_cat_idx)

        # Array (per-strain) of compartment indices for that strain's infectious compartments
        self._strain_infectious_indexers = {}
        # Array (per-strain) of len (strain infectious comps) that maps each
        # strain infectious compartment to its mixing category
        self._strain_category_indexers = {}
        #
        for strain in self.model._disease_strains:
            strain_filter = {"strain": strain} if "strain" in self.model.stratifications else {}
            strain_infectious_comps = self.model.query_compartments(
                strain_filter, tags="infectious", as_idx=True
            )
            self._strain_infectious_indexers[strain] = jnp.array(strain_infectious_comps)

            # Function that tests each element of an ndarray to see if is contained within the
            # current strain infectious compartments
            vcat = np.vectorize(lambda c: c in strain_infectious_comps)
            strain_cat_idx = pop_cat_idx[vcat(pop_cat_idx)]
            # Localize the compartment indices to be within strain infectious comps,
            # rather than the global compartment lookup
            # We do this by creating an inverse global map (but only fill in the compartment
            # indices we care about)
            # It's just a bit easier to read that messing around with dictionaries etc
            tlookup = np.empty(len(self.model.compartments), dtype=int)
            tlookup[strain_infectious_comps] = range(len(strain_infectious_comps))
            strain_cat_idx = tlookup[strain_cat_idx]

            # Ensure this is a 2d array
            # This necessary where there is only one compartment for each category
            strain_cat_idx = strain_cat_idx.reshape((ncats, int(strain_cat_idx.size / ncats)))
            self._strain_category_indexers[strain] = jnp.array(strain_cat_idx)

    def _precompute_flow_maps(self):
        """Build fast-access arrays of flow indices"""
        f_pos_map = []
        f_neg_map = []
        for i, f in self._iter_non_function_flows:
            if f.source:
                f_neg_map.append((i, f.source.idx))
            if f.dest:
                f_pos_map.append((i, f.dest.idx))

        self._pos_flow_map = np.array(f_pos_map, dtype=int)
        self._neg_flow_map = np.array(f_neg_map, dtype=int)

    def _build_infectious_multipliers_lookup(self):
        """Get multipliers for all infectious flows

        These are used by _get_infectious_multipliers_flat (currently experimental)

        Returns:
            np.ndarray: Array of infectiousness multipliers
        """
        lookups = []

        has_freq = False
        has_dens = False

        for i, idx in enumerate(self.infectious_flow_indices):
            f = self.model.flows[idx]
            if isinstance(f, flows.InfectionFrequencyFlow):
                has_freq = True
            elif isinstance(f, flows.InfectionDensityFlow):
                has_dens = True
            cat_idx, strain = self._get_infection_multiplier_indices(f.source, f.dest)
            strain_idx = self.model._disease_strains.index(strain)
            lookups.append([strain_idx, cat_idx])
        full_table = np.array(lookups, dtype=int)
        self._full_table = full_table.reshape(len(self.infectious_flow_indices), 2)
        self._infect_strain_lookup_idx = self._full_table[:, 0].flatten()
        self._infect_cat_lookup_idx = self._full_table[:, 1].flatten()

        self._infection_frequency_only = False
        self._infection_density_only = False

        self._infection_process_type = None

        if has_freq:
            if has_dens:
                self._infection_process_type = "both"
            else:
                self._infection_process_type = "freq"
                self._infection_frequency_only = True
        elif has_dens:
            self._infection_density_only = True
            self._infection_process_type = "dens"

    def _get_force_idx(self, source: Compartment):
        """
        Returns the index of the source compartment in the infection multiplier vector.
        """
        return self._category_lookup[source.idx]

    def _get_infection_multiplier_indices(
        self, source: Compartment, dest: Compartment
    ) -> Tuple[str, int]:
        """Return indices for infection frequency lookups"""
        idx = self._get_force_idx(source)
        strain = dest.strata.get("strain", self.model._DEFAULT_DISEASE_STRAIN)
        return idx, strain



"""
General utility functions used accross submodules
"""

import pandas as pd
from datetime import timedelta, datetime
from typing import Iterable
import numpy as np


def get_scenario_start_index(base_times, start_time):
    """
    Returns the index of the closest time step that is at, or before the scenario start time.
    """
    assert (
        base_times[0] <= start_time
    ), f"Scenario start time {start_time} is before baseline has started"
    indices_after_start_index = [idx for idx, time in enumerate(base_times) if time > start_time]
    if not indices_after_start_index:
        raise ValueError(f"Scenario start time {start_time} is set after the baseline time range")

    index_after_start_index = min(indices_after_start_index)
    start_index = max([0, index_after_start_index - 1])
    return start_index


def ref_times_to_dti(ref_date: datetime, times: Iterable) -> pd.DatetimeIndex:
    """Return a DatetimeIndex whose values after ref_date offset by times

    Args:
        ref_date (datetime): The reference  date ('epoch') from which offsets are computed
        times (Iterable): Offsets to ref_date (int or float)

    Returns:
        (pd.DatetimeIndex): Index suitable for constructing Series or DataFrame
    """

    return pd.DatetimeIndex([ref_date + timedelta(t) for t in times])


def clean_compartment_values(compartment_values: np.ndarray):
    """
    Zero out -ve compartment sizes in flow rate calculations,
    to prevent negative values from messing up the direction of flows.
    We don't expect large -ve values, but there can be small ones due to numerical errors.
    """
    comp_vals = compartment_values.copy()
    zero_mask = comp_vals < 0
    comp_vals[zero_mask] = 0
    return comp_vals


from numbers import Real


class Epoch:
    """Epoch converts between numeric offset values (eg model times), and concrete datetime values"""

    def __init__(self, ref_date: datetime, unit: timedelta = timedelta(1)):
        """Create an Epoch for conversion. ref_date will be equivalent to 0.0,
        with other dates offset in units of unit

        Args:
            ref_date: The '0-time' reference data
            unit (optional): timedelta unit equivalent to 1.0 steps in numeric representation
        """
        self.ref_date = ref_date
        self.unit = unit

    def __repr__(self):
        return f"Epoch from {self.ref_date}, in units of {self.unit}"

    def index_to_dti(self, index: pd.Index) -> pd.DatetimeIndex:
        """Convert an index (or iterable) of numeric values to a DatetimeIndex

        Args:
            index (pd.Index): Inputs to convert

        Returns:
            Equivalent DatetimeIndex
        """
        if not isinstance(index, pd.Index):
            index = pd.Index(index)
        return self.ref_date + index * self.unit

    def dti_to_index(self, index: pd.DatetimeIndex) -> pd.Index:
        """Convert a DatetimeIndex to its equivalent numeric representation

        Args:
            index: The input index to convert

        Returns:
            Equivalent numerical index
        """
        return (index - self.ref_date) / self.unit

    def number_to_datetime(self, n: Real) -> datetime:
        """Convert a single number to a datetime

        Args:
            n: The number

        Returns:
            The datetime
        """
        return self.ref_date + n * self.unit

    def datetime_to_number(self, d: datetime) -> float:
        """Convert a datetime number to a float

        Args:
            d: The datetime

        Returns:
            The number
        """
        return (d - self.ref_date) / self.unit



"""Builder functions for interpolating data - replaces scale_up.py

Note: For most user applications, refer to the get_*_function wrappers in time.py

build_static_sigmoidal_multicurve is the equivalent of scale_up_function(method=4),
with build_sigmoidal_multicurve producing a dynamic (parameterizable) version of this

"""

from jax import Array, lax, numpy as jnp

from .util import binary_search_ge


def _uncorrected_sigmoid(x: float, curvature: float) -> float:
    """Return a sigmoid of x (assumed to be between 0.0 and 1.0),
    whose shape and range depends on curvature

    Args:
        x: Input to the function
        curvature: Degree of sigmoidal flattening - 1.0 is linear, higher values increase smoothing

    Returns:
        Output of sigmoidal curve function
    """
    arg = curvature * (0.5 - x)
    return 1.0 / (1.0 + jnp.exp(arg))


def make_norm_sigmoid(curvature: float) -> callable:
    """
    Build a sigmoid function with fixed curvature, whose output is normalized to (0.0,1.0) across
    the (0.0,1.0) input range
    Args:
        curvature: See _uncorrected_sigmoid

    Returns:
        The normalized sigmoid function
    """
    offset = _uncorrected_sigmoid(0.0, curvature)
    scale = 1.0 / (1.0 - (offset * 2.0))

    def sig(x):
        return (_uncorrected_sigmoid(x, curvature) - offset) * scale

    return sig


def build_sigmoidal_multicurve(x_points: Array, curvature=16.0) -> callable:
    """Build a sigmoidal smoothing function across points specified by
    x_points; the returned function takes the y values as arguments in
    form of a scale_data dict with keys [min, max, values, ranges]

    Args:
        x_points: x values to interpolate across
        curvature: Sigmoidal curvature.  Default produces the same behaviour as the old
                   scale_up_function

    Returns:
        The multisigmoidal function of (x, scale_data)
    """
    xranges = jnp.diff(jnp.array(x_points))
    x_points = jnp.array(x_points)

    xmin = x_points.min()
    xmax = x_points.max()
    xbounds = jnp.array([xmin, xmax])

    sig = make_norm_sigmoid(curvature)

    def get_curve_at_t(t, values, ranges):
        # idx = sum(t >= x_points) - 1
        idx = binary_search_ge(t, x_points)

        offset = t - x_points[idx]
        relx = offset / xranges[idx]
        rely = sig(relx)
        return values[idx] + (rely * ranges[idx])

    def scaled_curve(t: float, ydata: dict):
        # Branch on whether t is in bounds
        bounds_state = sum(t > xbounds)
        branches = [
            lambda _, __, ___: ydata["min"],
            get_curve_at_t,
            lambda _, __, ___: ydata["max"],
        ]
        return lax.switch(bounds_state, branches, t, ydata["values"], ydata["ranges"])

    return scaled_curve


def build_linear_interpolator(x_points: Array) -> callable:
    """Build a sigmoidal smoothing function across points specified by
    x_points; the returned function takes the y values as arguments in
    form of a scale_data dict with keys [min, max, values, ranges]

    Args:
        x_points: x values to interpolate across
        curvature: Sigmoidal curvature.  Default produces the same behaviour as the old
                   scale_up_function

    Returns:
        The multisigmoidal function of (x, scale_data)
    """
    xranges = jnp.diff(jnp.array(x_points))
    x_points = jnp.array(x_points)

    xmin = x_points.min()
    xmax = x_points.max()
    xbounds = jnp.array([xmin, xmax])

    def get_curve_at_t(t, values, ranges):
        # idx = sum(t >= x_points) - 1
        idx = binary_search_ge(t, x_points)

        offset = t - x_points[idx]
        relx = offset / xranges[idx]
        return values[idx] + (relx * ranges[idx])

    from jax import lax

    def linear(t: float, ydata: dict):
        # Branch on whether t is in bounds
        bounds_state = sum(t > xbounds)
        branches = [
            lambda _, __, ___: ydata["min"],
            get_curve_at_t,
            lambda _, __, ___: ydata["max"],
        ]
        return lax.switch(bounds_state, branches, t, ydata["values"], ydata["ranges"])

    return linear


def get_scale_data(points: Array) -> dict:
    """
    Precompute min, max, and ranges for a set of data to be used in a scaling function such as that
    produced by build_sigmoidal_multicurve.  The onus is on the caller of this function to ensure
    they are the length expected by the target callee
    """
    ranges = jnp.diff(points)
    ymin = points[0]
    ymax = points[-1]

    data = {"min": ymin, "max": ymax, "values": points, "ranges": ranges}

    return data



import numpy as np

from summer2 import CompartmentalModel
from summer2 import Stratification, StrainStratification
from summer2.parameters import Parameter as param
from summer2.parameters import DerivedOutput


def sir(times=[0, 100]):
    m = CompartmentalModel(times, ["S", "I", "R"], "I")
    m.set_initial_population({"S": 990.0, "I": 10.0})
    m.add_infection_frequency_flow("infection", param("contact_rate"), "S", "I")
    m.add_transition_flow("recovery", param("recovery_rate"), "I", "R")
    incidence = m.request_output_for_flow("incidence", "infection")
    m.request_function_output("notifications", incidence * param("cdr"))
    m.set_default_parameters({"contact_rate": 0.4, "recovery_rate": 0.1, "cdr": 0.2})
    return m


def sirs_parametric_age(times=[0, 100], agegroups=list(range(0, 80, 5))):
    m = CompartmentalModel(times, ["S", "I", "R"], "I")
    m.set_initial_population({"S": 99990.0, "I": 10.0})
    m.add_infection_frequency_flow("infection", param("contact_rate"), "S", "I")
    m.add_transition_flow("recovery", 1.0 / param("recovery_duration"), "I", "R")
    m.add_transition_flow("waning", 1.0 / param("waning_duration"), "R", "S")

    max_strl = len(str(agegroups[-1]))
    agegroup_keys = [str(k).zfill(max_strl) for k in agegroups]
    num_age = len(agegroup_keys)
    age_strat = Stratification("age", agegroup_keys, ["S", "I", "R"])

    # +++ SUMMER3
    # Not currently parameterizable
    rec_adj = {k: adj for k, adj in zip(agegroup_keys, np.linspace(1.5, 0.5, num_age))}
    age_strat.set_flow_adjustments("recovery", rec_adj)
    wane_adj = {k: adj for k, adj in zip(agegroup_keys, np.linspace(0.5, 1.5, num_age))}
    age_strat.set_flow_adjustments("waning", wane_adj)

    mm_base = np.linspace(1.5, 1.0, num_age).reshape((1, num_age))
    mm = (mm_base * mm_base.T) * 0.1

    age_strat.set_mixing_matrix(mm)

    pop_spread = np.linspace(2.0, 1.0, num_age)
    pop_split = pop_spread / pop_spread.sum()

    age_strat.set_population_split({k: pop_prop for k, pop_prop in zip(agegroup_keys, pop_split)})

    m.stratify_with(age_strat)

    # +++ SUMMER3
    # Arbitrary dims for output requests
    # Export to xarray instead of dataframe
    incidence = m.request_output_for_flow("incidence", "infection")
    m.request_function_output("notifications", incidence * param("cdr"))

    for k in agegroup_keys:
        m.request_output_for_flow(f"incidenceXage_{k}", "infection", source_strata={"age": k})

    m.set_default_parameters(
        {"contact_rate": 0.4, "recovery_duration": 10.0, "waning_duration": 100.0, "cdr": 0.2}
    )
    return m
